import pathlib
import subprocess
import typing as typ
from datetime import date, datetime, time, timedelta

from wdid.dt_util import AUCKLAND_TZ, UTC, parse_date_string
from wdid.entrys import CombinedEntry, Entry
from .local_history import PROJECT_CONTAINING_FOLDERS

FIVE_MINUTES = timedelta(minutes=5)


local_history_dir = pathlib.Path("~").expanduser() / ".netbeans/"

ATOM_TIME_FORMATS = ["%Y-%m-%d_%I-%M-%S %p", "%Y-%m-%d_%H-%M-%S"]


def parse_local_history(today: date) -> typ.Collection[Entry]:
    if not local_history_dir.is_dir():
        return []
    data = []
    for line in subprocess.check_output(
        ["find", "-name", today.strftime("%Y-%m-%d") + "*"],
        cwd=str(local_history_dir),
        universal_newlines=True,
    ).splitlines():
        folder, _, filename = line.rpartition("/")
        time_end = filename.index("_", 15)
        time = filename[:time_end]
        dt = parse_date_string(time, ATOM_TIME_FORMATS)
        filename = folder[1:] + "/" + filename[time_end + 1 :]
        data.append(Entry(dt, "local-history", filename))
    return data


def compact_data(data):
    compacted_data = []
    prev_project_path = ""
    prev_project_ts = datetime.fromtimestamp(0)
    prev_project_entries = []

    def _commit():
        nonlocal compacted_data, prev_project_path, prev_project_ts
        nonlocal prev_project_entries
        if len(prev_project_entries) >= 2:
            compacted_data.append(
                CombinedEntry.from_entries(
                    "local-history", prev_project_path, prev_project_entries
                )
            )
        else:
            compacted_data += prev_project_entries
        prev_project_path = ""
        prev_project_ts = datetime.fromtimestamp(0)
        prev_project_entries = []

    for entry in data:
        if entry.owner != "local-history":
            _commit()
            compacted_data.append(entry)
            continue
        proj_path = to_project_path(entry.entry)
        if proj_path != prev_project_path:
            _commit()
        if (prev_project_ts + FIVE_MINUTES) < entry.timestamp:
            _commit()
        prev_project_path = proj_path
        prev_project_ts = entry.timestamp
        prev_project_entries.append(entry)
    _commit()

    return compacted_data


def to_project_path(filename):
    fpath = pathlib.Path(filename)
    for path in PROJECT_CONTAINING_FOLDERS:
        try:
            project_name = fpath.relative_to(path).parts[0]
            return path / project_name
        except ValueError:
            pass
    return str(fpath)



import os
import subprocess
import typing as typ
from datetime import date, datetime, time

from wdid.dt_util import parse_date_string
from wdid.entrys import CombinedEntry, Entry

XCHAT_TIME_FORMAT = ("%b %d %H:%M:%S",)
XCHAT_USERNAME = "opal"

xchat_dirs = [
    os.path.expanduser("~/.config/hexchat/logs/"),
    os.path.expanduser("~/.xchat2/xchatlogs/"),
]


def output_lines(command: typ.Sequence[str], **kwargs: typ.Any) -> typ.Collection[str]:
    return subprocess.run(
        command, universal_newlines=True, check=True, stdout=subprocess.PIPE, **kwargs
    ).stdout.splitlines()


def parse_xchat_logs(today: date) -> typ.Collection[Entry]:
    xchat_date = today.strftime("%b %d")
    dirs = filter(os.path.exists, xchat_dirs)
    if not dirs:
        return []
    data: typ.List[Entry] = []
    try:
        for file in output_lines(
            [
                "grep",
                "--text",
                "-irl",
                xchat_date + " .* <" + XCHAT_USERNAME + ">",
                *dirs,
            ]
        ):
            _, _, channel = file.rpartition("/")
            channel, _, _ = channel.rpartition(".")
            for line in output_lines(["grep", "--text", "-i", xchat_date, file]):
                if "opal" not in line:
                    if (
                        " has quit (" in line
                        or ") has joined " + channel in line
                        or "chanserv gives channel operator status to " in line
                    ):
                        continue
                    if "#" in channel:
                        continue
                try:
                    dt = parse_date_string(line[:15], XCHAT_TIME_FORMAT)
                    dt = dt.replace(year=today.year)
                    e = Entry(timestamp=dt, owner=channel, entry=line[16:])
                    if e not in data:
                        data.append(e)
                except ValueError:
                    pass
    except subprocess.CalledProcessError:
        pass
    return data



import pathlib
import typing as typ
import xml.etree.ElementTree as ET
from datetime import date, datetime, time

from wdid.dt_util import AUCKLAND_TZ, UTC
from wdid.entrys import CombinedEntry, Entry
from wdid.fs_util import riterdir

polari_dir = pathlib.Path("~/.local/share/TpLogger/logs/").expanduser().resolve()


def parse_polari_chat_history(today: date) -> typ.Collection[Entry]:
    data: typ.List[Entry] = []
    for file in riterdir(polari_dir):
        if file.suffix == ".log":
            # .../<name>/<date>.log
            chatname = file.parts[-2]
            if chatname.lower() == "nickserv":
                continue
            room_data = []
            have_messaged = False
            try:
                root = ET.parse(str(file)).getroot()
            except ET.ParseError:
                continue
            for msgelm in root.iter("message"):
                time = datetime.strptime(msgelm.attrib["time"], "%Y%m%dT%H:%M:%S")
                time = (
                    time.replace(tzinfo=UTC)
                    .astimezone(AUCKLAND_TZ)
                    .replace(tzinfo=None)
                )
                if time.date() != today:
                    break
                msg = "".join(msgelm.itertext())
                sep = ":"
                if msgelm.attrib["type"] == "action":
                    sep = " "
                have_messaged = have_messaged or (
                    msgelm.attrib["name"] in ("opal", "opal_", "opal_nz")
                )
                msg = "{:<12}{:<3}{}".format(msgelm.attrib["name"], sep, msg)
                room_data.append(Entry(time, chatname, msg))
            if have_messaged:
                data += room_data
    return data



import pathlib
import re
import typing as typ
from datetime import date, datetime, time, timedelta
from functools import lru_cache

from wdid.bisect import bisect_matches_get_range
from wdid.dt_util import AUCKLAND_TZ, parse_date_string
from wdid.entrys import CombinedEntry, Entry

DATE_FORMATS = ("[%A, %d %B %Y]",)
TIME_FORMATS = ("[%I:%M:%S %p %Z]",)
logs_dir = pathlib.Path("~/logs/").expanduser().resolve()
DATE_RE = re.compile(r"^\[.*?\]")
REST_RE = re.compile("(\[.*?\]) (<.*>|Away)?\s+(.*)$")


def parse_line(line: str) -> typ.Tuple[typ.Optional[str], str]:
    match = DATE_RE.match(line)
    if not match:
        return None, line
    return match.group(0), line[match.endpos :]


def bisect_cmp(min_date: date, max_date: date, line: str) -> typ.Optional[int]:
    date_str, _ = parse_line(line)
    if not date_str:
        return None
    date = parse_date_string(date_str, DATE_FORMATS).date()
    if date < min_date:
        return -1
    elif date > max_date:
        return 1
    else:
        return 0


def parse_konversation_chat_history(today: date) -> typ.Collection[Entry]:
    min_date = today - timedelta(days=1)
    max_date = today + timedelta(days=1)
    data: typ.List[Entry] = []
    for file in logs_dir.iterdir():
        if file.suffix == ".log":
            stem = file.stem
            if not stem.startswith("catalyst_"):
                continue
            room = stem[len("catalyst_") :]

            room_data = []
            keep_room = room[0] != "#"
            with file.open("r") as f:
                lines = list(f)
                matched_lines = bisect_matches_get_range(
                    lines, lambda line: bisect_cmp(min_date, max_date, line)
                )
                for line in matched_lines:
                    date_str, rest = parse_line(line)
                    if not date_str:
                        continue
                    match = REST_RE.fullmatch(line)
                    if not match:
                        continue
                    time_str, user, message = match.groups()
                    time = (
                        datetime.combine(
                            parse_date_string(date_str, DATE_FORMATS).date(),
                            parse_date_string(time_str, TIME_FORMATS).time(),
                        )
                        .astimezone(AUCKLAND_TZ)
                        .replace(tzinfo=None)
                    )
                    if time.date() != today:
                        continue
                    if user == "Away":
                        if "no longer" in message:
                            user = "Back"
                        else:
                            time = time - timedelta(minutes=10)
                        room_data.append(Entry(time, user, ""))
                        continue
                    if user is not None:
                        message = "{:<12}{}".format(user, message)
                        if "opal" in user:
                            keep_room = True
                    room_data.append(Entry(time, room, message))
            if keep_room:
                data += room_data
    return data



import pathlib
import sqlite3
import typing as typ
from datetime import date, datetime, time

from wdid.entrys import CombinedEntry

DATABASE = pathlib.Path("~/.local/share/hamster-gtk/hamster-gtk.sqlite").expanduser()


ACTIVITIES_QUERY = """
SELECT activities.name as name,
       "start", "end"
  FROM facts LEFT JOIN activities ON (facts.activity_id = activities.id)
 WHERE "start" > ? AND "start" < ? AND "end" IS NOT NULL
 ORDER BY "start"
"""


def parse_datetime(dt_str: typ.Union[str, bytes]) -> datetime:
    if isinstance(dt_str, bytes):
        dt_str = dt_str.decode("utf-8")
    return datetime.strptime(dt_str, "%Y-%m-%d %H:%M:%S.%f")


sqlite3.register_converter("DATETIME", parse_datetime)
# sqlite3.register_converter(datetime, datetime.isoformat)


def parse_hamster(today: date) -> typ.Collection[CombinedEntry]:
    if not DATABASE.exists():
        return []
    conn = sqlite3.connect(
        "file:%s?mode=ro" % DATABASE, uri=True, detect_types=sqlite3.PARSE_COLNAMES
    )
    conn.row_factory = sqlite3.Row
    start = datetime.combine(today, time.min).isoformat(" ")
    end = (datetime.combine(today, time.max)).isoformat(" ")
    cur = conn.execute(ACTIVITIES_QUERY, (start, end))
    data = []
    for row in cur:
        data.append(
            CombinedEntry(
                timestamp=parse_datetime(row["start"]),
                owner="hamster",
                entry=row["name"],
                end_timestamp=parse_datetime(row["end"]),
                original_entries=(),
            )
        )
    return data



#!/usr/bin/env python3
import itertools
from datetime import date, datetime, timedelta
from functools import partial

import click

from .dt_util import (
    ensure_date,
    next_month_start,
    to_month_end,
    to_month_start,
    to_week_end,
    to_week_start,
    TODAY,
)
from .cli import create_cli, print_entries, attach_aliases, DATE_FORMATS, tks_entries
from .plugins import load_data_for_date
from .ts.__main__ import cli as ts_cli


@click.group()
@click.option("--summary/--no-summary", "-s/-S", "summary", default=True)
@click.option("--detail/--no-detail", "-d/-D", "detail", default=None)
@create_cli
def cli(ctx, summary, detail):
    ctx.obj["summary"] = summary
    ctx.obj["detail"] = detail


@cli.command()
@click.argument("date", default=TODAY, type=click.DateTime(formats=DATE_FORMATS))
@click.pass_context
def date(ctx, date):
    summary = ctx.obj["summary"]
    detail = ctx.obj["detail"]
    day = ensure_date(date)
    data = load_data_for_date(day)
    click.echo(tks_entries(day, data, summary=summary, detail=detail))


cli.add_command(ts_cli, name="ts")
attach_aliases(cli, date)

if __name__ == "__main__":
    cli()



import os
import pathlib
import subprocess
import typing as typ
from datetime import date, datetime, time, timedelta

from wdid.dt_util import AUCKLAND_TZ, UTC, parse_date_string
from wdid.entrys import CombinedEntry, Entry
from wdid.fs_util import riterdir

local_history_dirs = [
    pathlib.Path("~/.atom/local-history/").expanduser(),
    pathlib.Path("~/.vscode/.history/").expanduser(),
]

ATOM_TIME_FORMATS = ("%Y-%m-%d_%I-%M-%S %p", "%Y-%m-%d_%H-%M-%S")

PROJECT_CONTAINING_FOLDERS = []
for raw_path in [
    "~/Fairfax",
    "~/Scratchpad/Firefox-Dev/",
    "~/Scratchpad",
    "~/nativform",
    "~/catinternal",
    "~/piwik3/",
]:
    path = pathlib.Path(raw_path).expanduser()
    if path.is_dir():
        PROJECT_CONTAINING_FOLDERS.append(path.resolve())
    del path


def parse_local_history(today: date) -> typ.Collection[Entry]:
    root = pathlib.Path("/")
    data: typ.List[Entry] = []
    file_bases = {}
    for local_history_dir in local_history_dirs:
        if not local_history_dir.is_dir():
            return data
        prefix = today.strftime("%Y-%m-%d")
        for file in riterdir(local_history_dir):
            if not file.is_file() or not file.name.startswith(prefix):
                continue
            time_end = file.name.index("_", 15)
            time = file.name[:time_end]
            dt = parse_date_string(time, ATOM_TIME_FORMATS)
            if file.parent in file_bases:
                file_base = file_bases[file.parent]
            else:
                file_base = file_bases[file.parent] = root / file.parent.relative_to(
                    local_history_dir
                )
            filename = file_base / file.name[time_end + 1 :]
            data.append(Entry(dt, "local-history", to_project_path(filename)))
    print(f"LH done with {today}")
    return data


def to_project_path(fpath: pathlib.Path) -> str:
    for path in PROJECT_CONTAINING_FOLDERS:
        try:
            project_name = fpath.relative_to(path).parts[0]
            return str(path / project_name)
        except ValueError:
            pass
    return str(fpath)



import pathlib
import typing as typ
from functools import lru_cache


@lru_cache(maxsize=2 << 10)
def riterdir(
    path: pathlib.Path, include_dirs: bool = False
) -> typ.Iterator[pathlib.Path]:
    if not path.exists():
        return
    path = path.resolve()
    paths = []
    for subpath in path.iterdir():
        if subpath.is_file():
            paths.append(subpath)
        elif include_dirs and subpath.is_dir():
            paths.append(subpath)
        if subpath.is_dir() and not subpath.is_symlink():
            paths.extend(riterdir(subpath, include_dirs))
    return frozenset(paths)



import typing as typ
from datetime import datetime, timedelta

import attr

from .dt_util import QUARTER_HOUR, round_dt

FIVE_MINUTES = timedelta(minutes=5)


@attr.s(frozen=True)
class Entry:

    timestamp: datetime = attr.ib()
    owner: str = attr.ib()
    entry: str = attr.ib()

    def to_line_parts(self) -> typ.Tuple[str, str, str]:
        return (self.owner, self.timestamp.strftime("%H:%M"), self.entry)

    def to_line(self):
        return "{:<14}{:<20}{}".format(*self.to_line_parts())


@attr.s(frozen=True)
class CombinedEntry(Entry):

    end_timestamp: datetime = attr.ib()
    original_entries: typ.Collection[Entry] = attr.ib()

    @classmethod
    def from_entries(
        cls, owner: str, entry_name: str, original_entries: typ.Collection[Entry]
    ) -> "CombinedEntry":
        assert len(original_entries) >= 2
        sorted_entries = sorted(original_entries)
        first_ts = sorted_entries[0].timestamp
        last_ts = sorted_entries[-1].timestamp
        return cls(first_ts, owner, entry_name, last_ts, tuple(sorted_entries))

    def to_line_parts(self) -> typ.Tuple[str, str, str]:
        et = self.end_timestamp.strftime("%H:%M")
        if self.timestamp != self.end_timestamp and et == "00:00":
            et = "24:00"
        ts = self.timestamp.strftime("%H:%M")
        return self.owner, f"{ts}-{et}", self.entry


def round_combined(
    entries: typ.Iterator[Entry], resolution: timedelta = QUARTER_HOUR
) -> typ.Iterator[Entry]:
    round_type = "down"
    for entry in entries:
        if isinstance(entry, CombinedEntry):
            start = round_dt(entry.timestamp, resolution, round_type)
            end = round_dt(entry.end_timestamp, resolution)
            yield attr.evolve(entry, timestamp=start, end_timestamp=end)
            round_type = "half_up"
        else:
            yield entry


def combine_close_entrys(
    data: typ.Collection[Entry]
) -> typ.Mapping[typ.Tuple[datetime, datetime], typ.Collection[Entry]]:
    grouping_len = timedelta(minutes=15)
    last_end = datetime.fromtimestamp(0)
    prev_data: typ.List[Entry] = []
    grouped = {}
    for entry in data:
        ts = entry.timestamp
        if entry.owner in ("Away", "Back") or ((last_end + grouping_len) < ts):
            if prev_data:
                first = prev_data[0].timestamp
                transformed = compact_data(prev_data)
                grouped[(first, last_end)] = tuple(transformed)
            prev_data = []
        if entry.owner not in ("#sysadmin", "#catalyst"):
            last_end = ts
        prev_data.append(entry)
    if prev_data:
        first = prev_data[0].timestamp
        transformed = compact_data(prev_data)
        grouped[(first, last_end)] = tuple(transformed)
    return grouped


def group_data(data: typ.Collection[Entry]) -> typ.Collection[Entry]:
    grouped_data = sorted(combine_close_entrys(data).items())
    out_data: typ.Iterator[Entry] = (
        CombinedEntry(start, "grouped", "🤷", end, ())
        for (start, end), _ in grouped_data
    )
    if round:
        out_data = round_combined(out_data)
    return sorted(out_data)


def compact_data(data: typ.Collection[Entry]) -> typ.Collection[Entry]:
    compacted_data: typ.List[Entry] = []
    prev_entry = ("", "")
    prev_ts = datetime.fromtimestamp(0)
    prev_entries: typ.List[Entry] = []

    def _commit() -> None:
        nonlocal prev_entry, prev_ts, prev_entries
        if len(prev_entries) <= 1:
            compacted_data.extend(prev_entries)
        else:
            new_entry = CombinedEntry.from_entries(
                prev_entry[0], prev_entry[1], prev_entries
            )
            compacted_data.append(new_entry)
        prev_entry = ("", "")
        prev_ts = datetime.fromtimestamp(0)
        prev_entries = []

    for entry in data:
        if (entry.owner, entry.entry) != prev_entry:
            _commit()
        elif (prev_ts + FIVE_MINUTES) < entry.timestamp:
            _commit()
        prev_entry = (entry.owner, entry.entry)
        prev_ts = entry.timestamp
        prev_entries.append(entry)
    _commit()

    return compacted_data



import typing as typ
from datetime import date, datetime, timedelta
from functools import partial

import click

from .dt_util import (
    TODAY,
    ensure_date,
    to_month_end,
    to_month_start,
    to_smart_month_start,
    to_smart_week_start,
    to_week_end,
    to_week_start,
)
from .entrys import Entry, compact_data, group_data
from .plugins import load_data_for_all_dates

DATE_FORMATS = ("%Y-%m-%d", "%Y/%m/%d", "%d-%m-%Y", "%d/%m/%Y")


def create_cli(sub_fn=None):
    @click.option(
        "--round-summary/--noround-summary", "-r/-R", "round_summary", default=True
    )
    @click.option("--color/--no-color", "-c/-C", "color", default=None)
    @click.pass_context
    def cli(ctx, round_summary, color, **kwargs):
        ctx.ensure_object(dict)
        if color in [True, False]:
            ctx.color = color
        if sub_fn:
            sub_fn(ctx, **kwargs)
        else:
            assert kwargs == {}

    if not sub_fn:
        return click.group()(cli)
    return cli


def create_ranged(sub_fn):
    month_start = to_smart_month_start(TODAY)

    @click.argument(
        "start", default=month_start, type=click.DateTime(formats=DATE_FORMATS)
    )
    @click.argument(
        "end",
        default=to_month_end(month_start),
        type=click.DateTime(formats=DATE_FORMATS),
    )
    @click.option("--inclusive/--exclusive", "-i/-e", "inclusive", default=False)
    @click.pass_context
    def ranged(ctx, start, end, inclusive, **kwargs):
        summary = ctx.obj["summary"]
        detail = ctx.obj["detail"]
        if detail is None:
            detail = False
        if inclusive:
            end += timedelta(days=1)
        start, end = ensure_date(start), ensure_date(end)
        days = [
            start + timedelta(days=day_offset)
            for day_offset in range((end - start).days)
        ]
        progress_load_data(days)
        sub_fn(ctx, days, **kwargs)

    return ranged


def print_entries(*a, **k):
    click.echo(stringify_entries(*a, **k))


def stringify_entries(
    entries: typ.Collection[Entry],
    indent: str = "  ",
    colors: typ.Union[None, bool, str, typ.Tuple[str, str, str]] = True,
):
    output = ""
    if colors is True:
        name_color, ts_color, val_color = ("green", "blue", "")
    if not colors or isinstance(colors, bool):  # isinstance to make mypy happy
        name_color, ts_color, val_color = ("", "", "")
    elif isinstance(colors, str):
        name_color, ts_color, val_color = colors, colors, colors
    else:
        name_color, ts_color, val_color = colors
    for entry in sorted(entries):
        name, ts, val = entry.to_line_parts()
        output += (
            indent
            + click.style(f"{name:<14}", fg=name_color)
            + click.style(f"{ts:<20}", fg=ts_color)
            + click.style(f"{val}", fg=val_color)
            + "\n"
        )
    return output.rstrip("\n")


def progress_load_data(days: typ.Collection[date]) -> typ.List[typ.Collection[Entry]]:
    with click.progressbar(label="Pre-loading data", length=len(days) + 1) as bar:
        bar.update(1)
        data = []
        for idx, dat in enumerate(load_data_for_all_dates(days)):
            bar.update(idx + 1)
            data.append(dat)
        return data


def tks_entries(
    day: date,
    data: typ.Collection[Entry],
    *,
    summary: bool,
    detail: bool,
    compact_detail: bool = True,
) -> str:
    output = ""
    if summary or detail:
        output += (
            click.style(day.strftime("%Y-%m-%d"), fg="green", bold=True)
            + " # "
            + click.style(day.strftime("%A"), fg="cyan")
            + "\n"
        )
    if summary:
        grouped = group_data(data)
        if not grouped:
            output += "# " + click.style("Day blank", fg="yellow") + "\n"
        else:
            output += stringify_entries(grouped, indent="") + "\n"
    if detail:
        if summary:
            output += click.style(("#" * 40) + "\n", dim=True) + "\n"
        if compact_detail:
            output += stringify_entries(compact_data(data)) + "\n"
        else:
            output += stringify_entries(data) + "\n"
    return output.rstrip("\n")


def attach_aliases(cli, single, ranged=None):  # typing: ignore
    if ranged is None:

        def simple_ranged_impl():
            progress_load_data(days)
            for day in days:
                ctx.forward(single, date=day)

        ranged = cli.command(name="range")(simple_ranged_impl)

    @cli.command()
    @click.pass_context
    def today(ctx):
        ctx.forward(single, date=TODAY)

    @cli.command()
    @click.pass_context
    def yesterday(ctx):
        ctx.forward(single, date=(TODAY - timedelta(days=1)))

    @cli.command()
    @click.pass_context
    def week(ctx):
        week_start = to_smart_week_start(TODAY)
        ctx.forward(
            ranged, start=week_start, end=to_week_end(week_start), inclusive=True
        )

    @cli.command()
    @click.pass_context
    def this_week(ctx):
        ctx.forward(
            ranged, start=to_week_start(TODAY), end=to_week_end(TODAY), inclusive=True
        )

    @cli.command()
    @click.pass_context
    def last_week(ctx):
        week_start = to_week_start(TODAY - timedelta(days=7))
        ctx.forward(
            ranged, start=week_start, end=to_week_end(week_start), inclusive=True
        )

    @cli.command()
    @click.pass_context
    def month(ctx):
        month_start = to_smart_month_start(TODAY)
        ctx.forward(
            ranged, start=month_start, end=to_month_end(month_start), inclusive=True
        )

    @cli.command()
    @click.pass_context
    def this_month(ctx):
        ctx.forward(
            ranged, start=to_month_start(TODAY), end=to_month_end(TODAY), inclusive=True
        )

    @cli.command()
    @click.pass_context
    def last_month(ctx):
        month_end = to_month_start(TODAY) - timedelta(days=1)
        ctx.forward(
            in_range, start=to_month_start(month_end), end=month_end, inclusive=True
        )



import itertools
import subprocess
import typing as typ
from datetime import date, datetime, timedelta
from multiprocessing import TimeoutError
from multiprocessing.pool import AsyncResult, Pool

from wdid.dt_util import Dateish, ensure_date
from wdid.entrys import Entry as _Entry  # Use `_` to prevent it's use in plugins

loaded_data = {}


def load_data_for_date(dateish: Dateish) -> typ.Collection[_Entry]:
    date = ensure_date(dateish)
    if date not in loaded_data:
        data = _load_data_for_date(date)
        loaded_data[date] = data
        return data
    else:
        return loaded_data[date]


def _load_data_for_date(date: date) -> typ.Collection[_Entry]:
    from .impl import hamster, hexchat, konversation, local_history, polari, zsh

    data: typ.List[_Entry] = []
    data += local_history.parse_local_history(date)
    data += hexchat.parse_xchat_logs(date)
    data += polari.parse_polari_chat_history(date)
    data += zsh.parse_zsh_history(date)
    data += konversation.parse_konversation_chat_history(date)
    data += hamster.parse_hamster(date)
    data = sorted(data)
    return tuple(data)


def _mp_pool_load(date: date):
    return date, _load_data_for_date(date)


def load_data_for_all_dates(
    dates: typ.Iterable[date]
) -> typ.Iterable[typ.Collection[_Entry]]:
    days = list(dates)
    del dates
    uniq_days = set(days)
    with Pool(processes=8) as pool:
        missing_days = set(filter(lambda day: day not in loaded_data, uniq_days))
        missing_iter = pool.imap_unordered(_mp_pool_load, missing_days)
        for day in days:
            if day in missing_days:
                for l_day, l_date in missing_iter:
                    loaded_data[l_day] = l_date
                    missing_days.discard(l_day)
                    if l_day == day:
                        break
            yield loaded_data[day]



import typing as typ
from logging import Logger
from typing import Generic, TypeVar

T = typ.TypeVar("T")


def bisect_matches_range(
    array: typ.Sequence[T], cmp: typ.Callable[[T], typ.Optional[int]]
) -> typ.Tuple[int, int]:
    start = bisect_matches_left(array, cmp)
    return start, bisect_matches_right(array, cmp, lo=start)


def bisect_matches_get_range(
    array: typ.Sequence[T], cmp: typ.Callable[[T], typ.Optional[int]]
) -> typ.Sequence[T]:
    start, end = bisect_matches_range(array, cmp)
    return array[start:end]


def bisect_matches_right(
    array: typ.Sequence[T],
    cmp: typ.Callable[[T], typ.Optional[int]],
    lo: int = 0,
    hi: typ.Optional[int] = None,
) -> int:
    """Return the index where to insert item x in list a, assuming a is sorted.
    The return value i is such that all e in a[:i] have e <= x, and all e in
    a[i:] have e > x.  So if x already appears in the list, a.insert(x) will
    insert just after the rightmost x already there.
    Optional args lo (default 0) and hi (default len(a)) bound the
    slice of a to be searched.
    """
    if hi is None:
        high = len(array)
    else:
        high = hi
    low, high = min((lo, high), (high, lo))
    low = min(max(0, low), len(array))
    high = min(max(0, high), len(array))
    del lo, hi
    while low < high:
        mid = (low + high) // 2
        cmp_res = cmp(array[mid])
        if cmp_res is None:
            low_test = bisect_matches_right(array, cmp, lo=low, hi=mid - 1)
            if low_test == mid - 1:
                return bisect_matches_right(array, cmp, lo=mid + 1, hi=high)
            else:
                return low_test
        elif cmp_res < 0:
            high = mid
        else:
            low = mid + 1
    return low


def bisect_matches_left(
    array: typ.Sequence[T],
    cmp: typ.Callable[[T], typ.Optional[int]],
    lo: int = 0,
    hi: typ.Optional[int] = None,
) -> int:
    """Return the index where to insert item x in list a, assuming a is sorted.
    The return value i is such that all e in a[:i] have e <= x, and all e in
    a[i:] have e > x.  So if x already appears in the list, a.insert(x) will
    insert just after the rightmost x already there.
    Optional args lo (default 0) and hi (default len(a)) bound the
    slice of a to be searched.
    """
    if hi is None:
        high = len(array)
    else:
        high = hi
    low, high = min((lo, high), (high, lo))
    low = min(max(0, low), len(array))
    high = min(max(0, high), len(array))
    del lo, hi
    while low < high:
        mid = (low + high) // 2
        cmp_res = cmp(array[mid])
        if cmp_res is None:
            low_test = bisect_matches_left(array, cmp, lo=mid + 1, hi=high)
            if low_test == mid + 1:
                return bisect_matches_left(array, cmp, lo=low, hi=mid - 1)
            else:
                return low_test
        elif cmp_res < 0:
            low = mid + 1
        else:
            high = mid
    return low



#!/usr/bin/env python3
import codecs
import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
about = {}  # type: Dict[str, Any]


with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()


with open(os.path.join(here, "wdid", "__version__.py")) as v:
    exec(v.read(), about)  # noqa: S102


required = ["pytz", "click", "attrs"]

setup(
    name="wdid",
    version=about["__version__"],
    description="Recall what you did on a certian day.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Opal Symes",
    author_email="opal@catalyst.net.nz",
    python_requires=">=3.6",
    install_requires=required,
    include_package_data=True,
    packages=find_packages(),
    entry_points={
        "console_scripts": ["wdid=wdid.__main__:cli", "wdid-ts=wdid.ts.__main__:cli"]
    },
    license="GPLv3+",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Topic :: Office/Business",
        "Topic :: Utilities",
    ],
)



import pathlib
import re
import typing as typ
from datetime import date, datetime, time

from wdid.dt_util import AUCKLAND_TZ, UTC
from wdid.entrys import CombinedEntry, Entry

zsh_history = pathlib.Path("~/.zsh_history").expanduser().resolve()


def parse_zsh_history(today: date) -> typ.Collection[Entry]:
    data: typ.List[Entry] = []
    with zsh_history.open("rb") as hist:
        item = None
        for line in hist:
            match = re.match(br"^: (\d+):\d;(.*)$", line)
            if match:
                if item:
                    time, cmd = item
                    item = None
                    data.append(Entry(time, "zsh", str(cmd)))
                timestamp, cmd = match.group(1, 2)
                time = datetime.fromtimestamp(int(timestamp), AUCKLAND_TZ)
                time = time.replace(tzinfo=None)
                if time.date() != today:
                    continue
                item = (time, cmd)
            else:
                if item:
                    item = (item[0], item[1] + line)
    return data



import typing as typ
from datetime import date, datetime, timedelta
from functools import lru_cache

from pytz import UTC, timezone

Dateish = typ.Union[date, datetime]

AUCKLAND_TZ = timezone("Pacific/Auckland")
QUARTER_HOUR = timedelta(minutes=15)
TODAY = datetime.now().date()


@lru_cache(maxsize=2 << 10)
def parse_date_string(date_string: str, formats: typ.Iterable[str]) -> datetime:
    for fmt in formats:
        try:
            return datetime.strptime(date_string, fmt)
            break
        except ValueError:
            pass
    raise ValueError("none of the formats matched the given date string.")


@lru_cache(maxsize=128)
def round_dt(dt: datetime, resolution, round_type="half_up") -> datetime:
    if isinstance(resolution, timedelta):
        resolution = resolution.total_seconds()
    tz = dt.tzinfo
    secs = dt.timestamp()
    diff = secs % resolution
    root = secs - diff
    res = root
    if round_type in ("down", "floor"):
        res = root
    elif round_type in ("up", "ceil", "ceiling"):
        if diff > 0:
            res = root + resolution
    elif round_type in ("half_up", "round"):
        if diff >= resolution / 2:
            res = root + resolution
    elif round_type == "half_down":
        if diff > resolution / 2:
            res = root + resolution
    else:
        raise ValueError("Unsupported round_type argument.")
    return datetime.fromtimestamp(res, tz=tz)


def ensure_date(dt: Dateish) -> date:
    if isinstance(dt, datetime):
        return dt.date()
    return dt


def to_month_start(dt: Dateish) -> date:
    return ensure_date(dt).replace(day=1)


def next_month_start(dt: Dateish) -> date:
    return to_month_start(to_month_start(dt) + timedelta(days=32))


def to_month_end(dt: Dateish) -> date:
    return next_month_start(dt) - timedelta(days=1)


def to_smart_month_start(dt: Dateish) -> date:
    d = ensure_date(dt)
    if d.day == 1:  # 1st of the month
        return to_month_start(d - timedelta(days=1))
    else:
        return to_month_start(d)


def to_week_start(dt: Dateish) -> date:
    # Week starts an Monday, and `weekday()` for monday == 0.
    d = ensure_date(dt)
    return d - timedelta(days=d.weekday())


def to_week_end(dt: Dateish) -> date:
    return to_week_start(dt) + timedelta(days=6)


def to_smart_week_start(dt: Dateish) -> date:
    d = ensure_date(dt)
    if d.weekday() == 0:
        return to_week_start(d - timedelta(days=1))
    else:
        return to_week_start(d)
